//  FluxboxEventCooker.cc for fluxbox.
//
//  Copyright (c) 2002 Steve Cooper, stevencooper@isomedia.com
//  Copyright (c) 1998-2000 by John Kennis, jkennis@chello.nl
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// (See the included file COPYING / GPL-2.0)
//

#include "FluxboxEvent.hh"
#include "BaseDisplay.hh"
#include <stdio.h>
#include <X11/Xmd.h>

FluxboxEventCooker::FluxboxEventCooker(
    BaseDisplay&               baseDisplay_,
    FluxboxCookedEventHandler& cookedHandler_)
  : baseDisplay(baseDisplay_),
    cookedHandler(cookedHandler_),
    initialized(false)
{
}

FluxboxEventCooker::~FluxboxEventCooker()
{
}

void FluxboxEventCooker::handleFluxboxEvents(XEvent& Event)
{
    if (Event.xclient.message_type != baseDisplay.getBlackboxStructureMessagesAtom())
        return;

    if (!initialized)
    {
        cookedHandler.onStartup();
        initialized = true;
    }

    unsigned event_atom = Event.xclient.data.l[0];
    if (event_atom == baseDisplay.getBlackboxNotifyWindowRaiseAtom())
    {
        cookedHandler.onWindowRaise(Event.xclient.data.l[1]);
    }
    else if (event_atom == baseDisplay.getBlackboxNotifyWindowLowerAtom())
    {
        cookedHandler.onWindowLower(Event.xclient.data.l[1]);
    }
    else if (event_atom == baseDisplay.getBlackboxNotifyWindowAddAtom())
    {
        cookedHandler.onWindowAdd(Event.xclient.data.l[1],Event.xclient.data.l[2]);
    }
    else if (event_atom == baseDisplay.getBlackboxNotifyWindowDelAtom())
    {
        cookedHandler.onWindowDelete(Event.xclient.data.l[1]);
    }
    else if (event_atom == baseDisplay.getBlackboxAttributesAtom())
    {
        cookedHandler.onWindowChangeAttributes(Event.xclient.data.l[1]);
    }
    else if (event_atom == baseDisplay.getBlackboxNotifyWindowFocusAtom())
    {
        cookedHandler.onWindowFocus(Event.xclient.data.l[1]);
    }
    if (event_atom == baseDisplay.getBlackboxNotifyCurrentWorkspaceAtom())
    {
        cookedHandler.onWorkspaceChangeCurrent(Event.xclient.data.l[1]);
    }
    else if (event_atom == baseDisplay.getBlackboxNotifyWorkspaceCountAtom())
    {
        cookedHandler.onWorkspaceChangeCount(Event.xclient.data.l[1]);
    }
}

void FluxboxEventCooker::handleCreateNotify(XEvent& event)
{
    cookedHandler.onWindowCreate(event.xcreatewindow.window, -1);
}

void FluxboxEventCooker::handleDestroyNotify(XEvent& event)
{
    cookedHandler.onWindowDestroy(event.xdestroywindow.window);
}

void FluxboxEventCooker::handleMapNotify(XEvent& event)
{
    cookedHandler.onWindowMap(event.xmap.window);
}

void FluxboxEventCooker::handleUnmapNotify(XEvent& event)
{
    cookedHandler.onWindowUnmap(event.xunmap.window);
}

void FluxboxEventCooker::handleReparentNotify(XEvent& event)
{
    cookedHandler.onWindowChangeParent(event.xreparent.window, event.xreparent.parent);
}

void FluxboxEventCooker::handleConfigureNotify(XEvent& event)
{
    cookedHandler.onWindowConfigure(event.xconfigure.window);
}

void FluxboxEventCooker::handlePropertyNotify(XEvent& event)
{
    XPropertyEvent& e = reinterpret_cast<XPropertyEvent&>(event);
    if (e.state != PropertyNewValue)
        return;
    if (e.atom == baseDisplay.getNETCurrentDesktopAtom())
    {
        long value;
        if (GetCardinalProperty(e.window, e.atom, 1, &value))
            cookedHandler.onWorkspaceChangeCurrent(value);
        return;
    }
}

bool FluxboxEventCooker::GetCardinalProperty(Window w, Atom& atom, long length, long* values)
{
    Atom actualType;
    int actualFormat;
    unsigned long actualLength, remaining;
    unsigned char *data;
    XGetWindowProperty(baseDisplay.getXDisplay(), w, atom, 0L, length, False, XA_CARDINAL,
                       &actualType, &actualFormat, &actualLength, &remaining, &data);
    if (data)
    {
        if (values != NULL)
        {
            long* pData = reinterpret_cast<long*>(data);
            for (unsigned long i = 0; i < static_cast<unsigned long>(length); i++)
            {
                if (i < actualLength)
                    values[i] = pData[i];
                else
                    values[i] = 0;
            }
        }
        XFree(data);
        return true;
    }
    else
        return false;
}
