// FluxspaceDisplay.cc for fluxspace - a workspace manager for fluxbox
//
//  Copyright (c) 2002 Steve Cooper, stevencooper@isomedia.com
//  Copyright (c) 1998-2000 by John Kennis, jkennis@chello.nl
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// (See the included file COPYING / GPL-2.0)
//

#include "FluxspaceDisplay.hh"
#include "FluxboxEvent.hh"
#include "../version.h"
#include <stdio.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

static int handleXErrors(Display *, XErrorEvent *);

FluxspaceDisplay::FluxspaceDisplay(const char* displayName, bool verbose_)
  : BaseDisplay("Fluxspace", displayName, verbose_),
    win(None),
    currentScreenInfo(NULL),
    verbose(verbose_),
    rawEventHandler(NULL)
{
    currentScreenInfo = getScreenInfo(DefaultScreen(getXDisplay()));
    XSetErrorHandler((XErrorHandler)handleXErrors);
    wmDeleteWindow = XInternAtom(getXDisplay(), "WM_DELETE_WINDOW",False);
    win = MakeWindow();
}

FluxspaceDisplay::~FluxspaceDisplay()
{
    if (win != None)
    {
        XUnmapWindow(getXDisplay(), win);
        XDestroyWindow(getXDisplay(), win);
    }
}

void FluxspaceDisplay::go(FluxboxRawEventHandler& rawEventHandler_)
{
    rawEventHandler = &rawEventHandler_;

    eventLoop();
}

void FluxspaceDisplay::reconfigure()
{
}

Window FluxspaceDisplay::MakeWindow()
{
    Window root = currentScreenInfo->getRootWindow();
    Window w = XCreateSimpleWindow(getXDisplay(), root, 0, 0, 52, 2, 0, 0, 0);

    XWMHints wmhints;
    memset(&wmhints, 0, sizeof wmhints);
    wmhints.initial_state = WithdrawnState;
    wmhints.flags = StateHint | InputHint;
    wmhints.input = False;
 
    XClassHint classhints;
    classhints.res_name=FLUXSPACE;
    classhints.res_class="fluxspace";
          
    char *name = FLUXSPACE;
    XTextProperty wname;
    XStringListToTextProperty(&name, 1, &wname);
    XSetWMProperties(getXDisplay(), w, &wname, NULL, NULL, 0, NULL, &wmhints, &classhints);

    Atom wmproto[2];
    wmproto[0] = wmDeleteWindow;
    wmproto[1] = getBlackboxStructureMessagesAtom();
    XSetWMProtocols(getXDisplay(), w, wmproto, 2);

    XMapWindow(getXDisplay(), w);

    return w;
}

void FluxspaceDisplay::process_event(XEvent* event)
{
    switch(event->type)
    {
      case CreateNotify:
        rawEventHandler->handleCreateNotify(*event);
        break;
      case DestroyNotify:
        rawEventHandler->handleDestroyNotify(*event);
        break;
      case MapNotify:
        rawEventHandler->handleMapNotify(*event);
        break;
      case UnmapNotify:
        rawEventHandler->handleUnmapNotify(*event);
        break;
      case ReparentNotify:
        rawEventHandler->handleReparentNotify(*event);
        break;
      case ConfigureNotify:
        rawEventHandler->handleConfigureNotify(*event);
        break;
      case PropertyNotify:
        rawEventHandler->handlePropertyNotify(*event);
        break;
      case ClientMessage:
        if ((unsigned)event->xclient.data.l[0] == wmDeleteWindow)
            shutdown();
        rawEventHandler->handleFluxboxEvents(*event);
        break;
    }
}

Bool FluxspaceDisplay::handleSignal(int sig)
{
    switch(sig)
    {
      case SIGHUP:
        break;

      case SIGSEGV:
      case SIGFPE:
      case SIGINT:
      case SIGTERM:
        shutdown();

      default:
        return False;
    }

    return True;
}

// X error handler to handle any and all X errors while blackbox is running
static int handleXErrors(Display *d, XErrorEvent *e)
{
  #ifdef DEBUG  
    char errtxt[128];
    XGetErrorText(d, e->error_code, errtxt, 128);
    fprintf(stderr,
            "fluxspace:  [ X Error event received. ]\n"
            "  X Error of failed request:  %d %s\n"
            "  Major/minor opcode of failed request:  %d / %d\n"
            "  Resource id in failed request:  0x%lx\n", e->error_code, errtxt,
            e->request_code, e->minor_code, e->resourceid);
  #endif
    return(False);
}
