#include <stdio.h>
#include "fluxspace.hh"
#include "FluxspaceDisplay.hh"
#include "PythonInterface.hh"
#include "FluxspaceInterface.hh"
#include "Esetroot.hh"

class ControllerImplementation
{
    // Used internally as singleton for Controller class
    friend class Controller;

  private:
    ControllerImplementation(const char* displayName, bool verbose_) throw(Error)
      : verbose(verbose_),
        startedEvents(false),
        display(NULL),
        fluxspace(NULL),
        pythonInterface(NULL),
        esetroot(NULL)
    {
        if (displayName == NULL)
            displayName = getenv("DISPLAY");
        if (displayName == NULL)
            throw ErrorNoDisplay();

        if (verbose)
            printf("display=%s\n", displayName);

        fluxspace = new FluxspaceDisplay(displayName, verbose);

        display = fluxspace->getXDisplay();
        if (display == NULL)
            throw ErrorNoDisplay(); // TODO: More specific exception?

        pythonInterface = new PythonInterface();

        esetroot = new Esetroot(display);
    }

    ~ControllerImplementation()
    {
        delete fluxspace;
        delete pythonInterface;
        delete esetroot;
    }

  public:
    bool              verbose;
    bool              startedEvents;
    Display*          display;
    FluxspaceDisplay* fluxspace;
    PythonInterface*  pythonInterface;
    Esetroot*         esetroot;
};

namespace
{
    static ControllerImplementation* imp = NULL;

    void CheckHealth() throw(Error)
    {
        if (   imp == NULL
#ifdef DEBUG
            || imp->display         == NULL
            || imp->fluxspace       == NULL
            || imp->pythonInterface == NULL
            || imp->esetroot        == NULL
#endif
           )
            throw ErrorBadController();
    }

    void CheckHealthAndWindow(Window window) throw(Error)
    {
        if (window == None)
            throw ErrorNoWindow();
        CheckHealth();
    }

    void GetProperty(
        Window      window,
        const Atom& propertyName,
        void*       buf,
        size_t      bufSize,
        const char* name) throw(Error)
    {
        CheckHealth();

        if (buf == NULL || bufSize <= 0)
            throw InternalError("Null or empty buffer provided to retrieve property");

        long offset = 0;
        long size32 = (bufSize + 3) / 4;
        Atom retType;
        int retFormat;
        unsigned long retCount, retRemaining;
        unsigned char* tmpPropBuf;

        if (XGetWindowProperty(imp->display,
                               window,
                               propertyName,
                               offset,
                               size32,
                               false,
                               AnyPropertyType,
                               &retType,
                               &retFormat,
                               &retCount,
                               &retRemaining,
                               &tmpPropBuf) != Success)
            throw ErrorWindowProperty(name, "get");

        size_t retPropSize = ( retFormat / 8 ) * retCount;
        if (tmpPropBuf == NULL || retPropSize <= 0)
        {
            memset(buf, 0, bufSize);
            return;
        }
        if (retPropSize > bufSize)
            throw InternalError("%ul byte buffer is too small for %ul byte \"%s\" property",
                                bufSize, retPropSize, name);

        memcpy(buf, tmpPropBuf, retPropSize);
        if (retPropSize < bufSize)
            memset(((char*)buf)+retPropSize, 0, bufSize - retPropSize);

        XFree(tmpPropBuf);
    }

    void SetProperty(
        Window      window,
        const Atom& propertyName,
        const Atom& propertyType,
        int         format,
        void*       buf,
        size_t      bufSize,
        const char* name) throw(Error)
    {
        CheckHealth();

        if (buf == NULL || bufSize <= 0)
            throw InternalError("Null or empty buffer provided to retrieve property");

        int nElements = bufSize / (format / 8);
        if (XChangeProperty(imp->display,
                            window,
                            propertyName,
                            propertyType,
                            format,
                            PropModeReplace,
                            (unsigned char*)buf,
                            nElements) != Success)
            throw ErrorWindowProperty(name, "set");
    }
};

Error::Error(const char* msg_, ...)
{
    char formattedMsg[1024];
    va_list args;
    va_start(args,msg_);
    vsnprintf(formattedMsg, sizeof formattedMsg, msg_, args);
    va_end(args);

    msg = formattedMsg;

    if (msg.length() > 0)
        printf("ERROR: %s\n", msg.c_str());
}

InternalError::InternalError(const char* msg_, ...)
  : Error("")
{
    char formattedMsg[1024];
    va_list args;
    va_start(args,msg_);
    vsnprintf(formattedMsg, sizeof formattedMsg, msg_, args);
    va_end(args);

    msg = formattedMsg;

    printf("INTERNAL ERROR: %s\n", msg.c_str());
}

Controller::Controller(const char* displayName, bool verbose_) throw(Error)
  : verbose(verbose_)
{
    if (imp != NULL)
        throw ErrorControllerExists();
    imp = new ControllerImplementation(displayName, verbose);
}

Controller::~Controller()
{
    delete imp;
    imp = NULL;
}

void Controller::RegisterFluxlet(const char* moduleName, PyObject* config) throw(Error)
{
    ::CheckHealth();
    imp->pythonInterface->AddModule(moduleName, config);
}

void Controller::StartEvents() throw(Error)
{
    ::CheckHealth();
    if (imp->startedEvents)
        throw ErrorEventLoopRunning();
    Window w = imp->fluxspace->getWindow(); // Hide probe window
    //Window w = None;                        // Don't hide probe window
    FluxspaceInterface fluxspaceInterface(*imp->fluxspace, *imp->pythonInterface, w, imp->verbose);
    FluxboxEventCooker fluxboxEventCooker(*imp->fluxspace, fluxspaceInterface);
    imp->startedEvents = true;
    imp->fluxspace->go(fluxboxEventCooker);
}

void Controller::Trace(const char* message)
{
    if (verbose && message != NULL)
        printf("fluxspace: %s\n", message);
}

void Controller::Message(int severity, const char* message)
{
    char s[50];
    s[0] = '\0';
    switch(severity)
    {
      case 1:
        strcpy(s, " WARNING");
        break;
      case 2:
        strcpy(s, " ERROR");
        break;
    }
    printf("fluxspace%s: %s\n", s, message);
}

FluxletHelper::FluxletHelper(const char* name_) throw(Error)
  : verbose(false)
{
    ::CheckHealth();
    verbose = imp->verbose;
    if (name_ != NULL)
        name = name_;
}

FluxletHelper::~FluxletHelper()
{
}

void FluxletHelper::SetRoot(
    const char* fname,
    const char* bgcolor,
    bool        scale,
    bool        center,
    bool        fit,
    bool        mirror) throw(Error)
{
    ::CheckHealth();
    imp->esetroot->SetRoot(fname, bgcolor, scale, center, fit, mirror);
}

void FluxletHelper::AddHandler(PyObject* object) throw(Error)
{
    ::CheckHealth();
    imp->pythonInterface->AddHandler(object);
}

void FluxletHelper::Trace(const char* message)
{
    if (verbose && message != NULL)
        printf("fluxlet.%s: %s\n", name.c_str(), message);
}

void FluxletHelper::Message(int severity, const char* message)
{
    char s[50];
    s[0] = '\0';
    switch(severity)
    {
      case 1:
        strcpy(s, " WARNING");
        break;
      case 2:
        strcpy(s, " ERROR");
        break;
    }
    printf("fluxlet.%s%s: %s\n", name.c_str(), s, message);
}
